package nxt.addons;

import nxt.util.Logger;

import java.lang.reflect.Field;
import java.util.Arrays;

public abstract class AbstractContract<InvocationData, ReturnedData> implements Contract<InvocationData, ReturnedData> {

    private JO params;

    public void setContractParams(JO params) {
        if (params == null) {
            throw new IllegalStateException("Cannot set null parameters to contract " + getClass().getCanonicalName());
        }
        this.params = params;
    }

    public JO getContractParams() {
        return params;
    }

    @Override
    public void processBlock(BlockContext context) {}

    @Override
    public void processTransaction(TransactionContext context) {
        JO jo = new JO();
        jo.put("errorDescription", String.format("Contract %s does not support processTransaction operation", getClass().getCanonicalName()));
        context.setResponse(jo);
    }

    @Override
    public void processRequest(RequestContext context) {
        JO jo = new JO();
        jo.put("contract", getClass().getCanonicalName());
        jo.put("account", context.getAccount());
        jo.put("accountRS", context.getAccountRs());
        jo.put("publicKey", context.getPublicKeyHexString());
        context.setResponse(jo);
    }

    @Override
    public void processVoucher(VoucherContext context) {
        JO jo = new JO();
        jo.put("errorDescription", String.format("Contract %s does not support processVoucher operation", getClass().getCanonicalName()));
        context.setResponse(jo);
    }

    @Override
    public ReturnedData processInvocation(DelegatedContext context, InvocationData data) {
        JO jo = new JO();
        jo.put("errorDescription", String.format("Contract %s does not support processInvocation operation", getClass().getCanonicalName()));
        context.setResponse(jo);
        return null;
    }

    public void injectParams(JO runnerConfigParams, JO contractSetupParams, JO invocationParams) {
        Arrays.stream(getClass().getDeclaredFields()).forEach(field -> {
            injectField(field, runnerConfigParams, ContractRunnerParameter.class);
            injectField(field, contractSetupParams, ContractSetupParameter.class);
            injectField(field, invocationParams, ContractInvocationParameter.class);
        });
    }

    private void injectField(Field field, JO params, Class parameterType) {
        if (field.getDeclaredAnnotation(parameterType) == null) {
            return;
        }
        if (!params.isExist(field.getName())) {
            return;
        }
        try {
            // Enable updating private field
            field.setAccessible(true);

            // Update the field value based on the supported data types
            if (field.getType().isAssignableFrom(boolean.class)) {
                field.set(this, params.getBoolean(field.getName()));
            } else if (field.getType().isAssignableFrom(int.class)) {
                field.set(this, params.getInt(field.getName()));
            } else if (field.getType().isAssignableFrom(long.class)) {
                if (field.getDeclaredAnnotation(BlockchainEntity.class) == null) {
                    field.set(this, params.getLong(field.getName()));
                } else {
                    field.set(this, params.getEntityId(field.getName()));
                }
            } else if (field.getType().isAssignableFrom(String.class)) {
                field.set(this, params.getString(field.getName()));
            } else if (field.getType().isAssignableFrom(JO.class)) {
                field.set(this, params.getJo(field.getName()));
            } else if (field.getType().isAssignableFrom(JA.class)) {
                field.set(this, params.getArray(field.getName()));
            } else {
                Logger.logWarningMessage(String.format("Param %s of type %s is not of one of the supported parameter types",
                        field.getName(), field.getType().getName()));
            }
        } catch (Throwable t) {
            Logger.logWarningMessage(String.format("Cannot update param %s of type %s error %s",
                    field.getName(), field.getType().getName(), t.getMessage()));
        }
    }

}
