package nxt.addons;

import nxt.Nxt;
import nxt.account.Account;
import nxt.blockchain.Chain;
import nxt.blockchain.ChildChain;
import nxt.crypto.Crypto;
import nxt.crypto.EncryptedData;
import nxt.util.Convert;
import nxt.util.Logger;

import java.util.HashMap;
import java.util.Map;

class ActiveContractRunnerConfig implements ContractRunnerConfig {

    private static final String ERROR_PREFIX = "contract runner config error: ";

    private final JO config;
    private final ContractProvider contractProvider;

    private String secretPhrase;
    private byte[] publicKey;
    private String publicKeyHexString;
    private long accountId;
    private String account;
    private String accountRs;
    private Map<Integer, Long> feeRatePerChain;
    private JO params;
    private boolean isValidator;
    private String validatorSecretPhrase;
    private int catchUpInterval;
    private byte[] runnerSeed;

    ActiveContractRunnerConfig(JO config, ContractProvider contractProvider) {
        this.config = config;
        this.contractProvider = contractProvider;
    }

    public void init() {
        initAccount();
        initFee();
        initParams();
        initValidation();
        initRandomSeed();
        initInterval();
    }

    private void initAccount() {
        secretPhrase = getProperty("secretPhrase");
        if (Convert.emptyToNull(secretPhrase) == null) {
            String accountRS = getProperty("accountRS");
            if (accountRS == null) {
                throw new IllegalArgumentException(ERROR_PREFIX + "secretPhrase or accountRS must be defined");
            }
            accountId = Convert.parseAccountId(accountRS);
            publicKey = Account.getPublicKey(accountId);
            if (publicKey == null) {
                throw new IllegalArgumentException(String.format(ERROR_PREFIX + "account %s does not have a public key", accountRS));
            }
        } else {
            accountId = Account.getId(Crypto.getPublicKey(secretPhrase));
            publicKey = Crypto.getPublicKey(secretPhrase);
        }
        publicKeyHexString = Convert.toHexString(publicKey);
        account = Long.toUnsignedString(accountId);
        accountRs = Convert.rsAccount(accountId);
    }

    private void initFee() {
        feeRatePerChain = new HashMap<>();
        for (Chain chain : ChildChain.getAll()) {
            String stringProperty = getProperty("feeRateNQTPerFXT." + chain.getName());
            if (stringProperty == null) {
                continue;
            }
            long fee = Long.parseLong(stringProperty);
            feeRatePerChain.put(chain.getId(), fee);
        }
        if (secretPhrase != null && feeRatePerChain.size() == 0) {
            throw new IllegalArgumentException(ERROR_PREFIX + "feeRateNQTPerFXT not specified for any chain");
        }
    }

    private void initParams() {
        // We do not load contract params from the properties file
        if (config.isExist("params")) {
            params = config.getJo("params");
        } else {
            params = new JO();
        }
    }

    private void initValidation() {
        isValidator = Boolean.parseBoolean(getProperty("validator"));
        if (isValidator) {
            validatorSecretPhrase = getProperty("validatorSecretPhrase");
            if (validatorSecretPhrase == null) {
                Logger.logWarningMessage("Contract runner validatorSecretPhrase not specified, contract won't be able to approve other contract transactions");
            } else if (Convert.emptyToNull(secretPhrase) != null) {
                throw new IllegalArgumentException(ERROR_PREFIX + "do not specify both secretPhrase and validatorSecretPhrase");
            }
        } else {
            validatorSecretPhrase = null;
        }
    }

    private void initRandomSeed() {
        String seed = getProperty("seed");
        if (seed != null) {
            runnerSeed = Convert.parseHexString(seed);
            if (runnerSeed.length < 16) {
                Logger.logWarningMessage("Contract runner random seed is shorter than 16 bytes, it might be possible to brute force it");
            }
        } else {
            Logger.logWarningMessage("Contract runner random seed not specified, random values generated by this contract runner will be predictable");
            runnerSeed = publicKey;
        }
    }

    private void initInterval() {
        catchUpInterval = config.getInt("catchUpInterval", 3600);
    }

    private String getProperty(String key) {
        if (config.isExist(key)) {
            return config.getString(key);
        } else {
            return Nxt.getStringProperty(ContractRunner.CONFIG_PROPERTY_PREFIX + key, null, key.toLowerCase().endsWith("secretphrase"));
        }
    }

    public String getSecretPhrase() {
        SecurityManager sm = System.getSecurityManager();
        if (sm != null) {
            sm.checkPermission(new ContractRunnerPermission("config"));
        }
        return secretPhrase;
    }

    public String getValidatorSecretPhrase() {
        SecurityManager sm = System.getSecurityManager();
        if (sm != null) {
            sm.checkPermission(new ContractRunnerPermission("config"));
        }
        return validatorSecretPhrase;
    }

    public byte[] getPublicKey() {
        return publicKey;
    }

    public String getPublicKeyHexString() {
        return publicKeyHexString;
    }

    public long getAccountId() {
        return accountId;
    }

    public String getAccount() {
        return account;
    }

    public String getAccountRs() {
        return accountRs;
    }

    public long getFeeRateNQTPerFXT(int chainId) {
        if (feeRatePerChain.get(chainId) == null) {
            return -1;
        }
        return feeRatePerChain.get(chainId);
    }

    public JO getParams() {
        return params;
    }

    public boolean isValidator() {
        return isValidator;
    }

    public int getCatchUpInterval() {
        return catchUpInterval;
    }

    public byte[] getRunnerSeed() {
        return runnerSeed;
    }

    public ContractProvider getContractProvider() {
        return contractProvider;
    }

    /**
     * Encrypt message sent to specific account
     * @param publicKey the recipient public key
     * @param data the message bytes
     * @param compress the compression mode
     * @return the encrypted data
     */
    public EncryptedData encryptTo(byte[] publicKey, byte[] data, boolean compress) {
        return Account.encryptTo(publicKey, data, secretPhrase, compress);
    }

    /**
     * Decrypt message sent to a specific account
     * @param publicKey the sender account public key
     * @param encryptedData the encrypted message object
     * @param uncompress the compression mode
     * @return the decrypted message bytes
     */
    public byte[] decryptFrom(byte[] publicKey, EncryptedData encryptedData, boolean uncompress) {
        return Account.decryptFrom(publicKey, encryptedData, secretPhrase, uncompress);
    }

    @Override
    public String getStatus() {
        return "Ok";
    }
}
