/*
 * Copyright © 2013-2016 The Nxt Core Developers.
 * Copyright © 2016-2018 Jelurida IP B.V.
 *
 * See the LICENSE.txt file at the top-level directory of this distribution
 * for licensing information.
 *
 * Unless otherwise agreed in a custom licensing agreement with Jelurida B.V.,
 * no part of this software, including this file, may be copied, modified,
 * propagated, or distributed except according to the terms contained in the
 * LICENSE.txt file.
 *
 * Removal or modification of this copyright notice is prohibited.
 *
 */

package nxt.addons;

import nxt.Constants;
import nxt.Nxt;
import nxt.NxtException;
import nxt.account.Account;
import nxt.blockchain.Block;
import nxt.blockchain.BlockchainProcessor;
import nxt.blockchain.Chain;
import nxt.blockchain.ChainTransactionId;
import nxt.blockchain.ChildBlockFxtTransactionType;
import nxt.blockchain.ChildChain;
import nxt.blockchain.ChildTransaction;
import nxt.blockchain.FxtChain;
import nxt.blockchain.Transaction;
import nxt.blockchain.TransactionProcessor;
import nxt.db.DbIterator;
import nxt.http.APICall;
import nxt.http.APIServlet;
import nxt.http.APITag;
import nxt.http.JSONData;
import nxt.http.responses.TransactionResponse;
import nxt.lightcontracts.ContractReference;
import nxt.messaging.PrunableEncryptedMessageAppendix;
import nxt.messaging.PrunablePlainMessageAppendix;
import nxt.util.Convert;
import nxt.util.Logger;
import nxt.util.ResourceLookup;
import nxt.voting.PhasingPollHome;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.JSONValue;

import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static nxt.addons.AbstractContractContext.EventSource;
import static nxt.addons.ContractRunner.INVOCATION_TYPE.BLOCK;
import static nxt.addons.ContractRunner.INVOCATION_TYPE.TRANSACTION;

public final class ContractRunner implements AddOn, ContractProvider {

    enum INVOCATION_TYPE {
        TRANSACTION, BLOCK, REQUEST, VOUCHER, INTERNAL_CALL
    }

    static final String CONFIG_PROPERTY_PREFIX = "addon.contractRunner.";
    private static final String CONFIG_FILE_PROPERTY = CONFIG_PROPERTY_PREFIX + "configFile";

    private ContractRunnerConfig config;
    private Map<String, Contract> supportedContracts = new HashMap<>();
    private Map<String, ContractReference> supportedContractReferences = new HashMap<>();
    private Map<String, APIServlet.APIRequestHandler> apiRequests;
    private Map<String, ContractReference> addedContractReferences = new HashMap<>();
    private Map<String, ContractReference> deletedContractReferences = new HashMap<>();

    @Override
    public void init() {
        // Initialize contract runner APIs
        SecurityManager sm = System.getSecurityManager();
        if (sm != null) {
            sm.checkPermission(new ContractRunnerPermission("init"));
        }
        apiRequests = new HashMap<>();
        apiRequests.put("getSupportedContracts", new ContractRunnerAPIs.GetSupportedContractsAPI(this, new APITag[]{APITag.ADDONS}, "adminPassword"));
        apiRequests.put("triggerContractByTransaction", new ContractRunnerAPIs.TriggerContractByTransactionAPI(this, new APITag[]{APITag.ADDONS}, "triggerFullHash", "apply", "validate", "adminPassword"));
        apiRequests.put("triggerContractByHeight", new ContractRunnerAPIs.TriggerContractByHeightAPI(this, new APITag[]{APITag.ADDONS}, "contractName", "height", "apply", "adminPassword"));
        apiRequests.put("triggerContractByRequest", new ContractRunnerAPIs.TriggerContractByRequestAPI(this, new APITag[]{APITag.ADDONS}, "contractName", "setupParams", "adminPassword"));
        apiRequests.put("triggerContractByVoucher", new ContractRunnerAPIs.TriggerContractByVoucherAPI(this, "voucher", new APITag[]{APITag.ADDONS}, "contractName", "adminPassword"));
        if (Constants.isTestnet && Constants.isAutomatedTest) {
            apiRequests.put("reloadContractRunnerConfiguration", new ContractRunnerAPIs.ReloadContractRunnerConfigurationAPI(this, new APITag[]{APITag.ADDONS}, "adminPassword", "configFileName"));
        }

        if (!Nxt.getServerStatus().isDatabaseReady()) {
            // For some utilities it is enough that we register the API even if Nxt itself is not initialized
            return;
        }

        // Read contract runner configuration
        loadConfig(Nxt.getStringProperty(CONFIG_FILE_PROPERTY));

        if (FxtChain.FXT.getBalanceHome().getBalance(config.getAccountId()).getUnconfirmedBalance() < Constants.UNCONFIRMED_POOL_DEPOSIT_FQT) {
            String message = String.format("contract runner account %s must have enough %s to pay the unconfirmed pool deposit", Convert.rsAccount(config.getAccountId()), FxtChain.FXT_NAME);
            config = new NullContractRunnerConfig(message);
            Logger.logErrorMessage(message);
            return;
        }
        try {
            try (DbIterator<ContractReference> iterator = ContractReference.getContractReferences(config.getAccountId(), null, 0, Integer.MAX_VALUE)) {
                while (iterator.hasNext()) {
                    ContractLoader.loadContract(iterator.next(), supportedContracts, supportedContractReferences);
                }
            }
        } catch (Throwable t) {
            String message = t.toString();
            config = new NullContractRunnerConfig(message);
            Logger.logErrorMessage(message);
            return;
        }
        // Register listeners for contract activation
        Nxt.getBlockchainProcessor().addListener(this::processBlock, BlockchainProcessor.Event.AFTER_BLOCK_ACCEPT);
        Nxt.getTransactionProcessor().addListener(this::processConfirmed, TransactionProcessor.Event.ADDED_CONFIRMED_TRANSACTIONS);
        Nxt.getTransactionProcessor().addListener(this::processReleasedPhased, TransactionProcessor.Event.RELEASE_PHASED_TRANSACTION);
        ContractReference.addListener(this::contractAdded, ContractReference.Event.SET_CONTRACT_REFERENCE);
        ContractReference.addListener(this::contractDeleted, ContractReference.Event.DELETE_CONTRACT_REFERENCE);
        Logger.logInfoMessage("ContractRunner Started");
    }

    void loadConfig(String configFileName) {
        JO configJson;
        if (configFileName != null) {
            configJson = AccessController.doPrivileged((PrivilegedAction<JO>) () -> ResourceLookup.loadJsonResource(configFileName));
            if (configJson == null) {
                Logger.logInfoMessage("Cannot load contract runner config from " + configFileName);
                configJson = new JO();
            }
        } else {
            configJson = new JO();
        }
        config = new ActiveContractRunnerConfig(configJson, this);
        config.init();
    }

    <T extends AbstractContractContext> void process(Contract contract, T context, INVOCATION_TYPE invocationType) {
        JO invocationParams = context instanceof AbstractOperationContext ? ((AbstractOperationContext) context).getRuntimeParams() : new JO();
        if (Constants.isAutomatedTest) {
            AccessController.doPrivileged((PrivilegedAction<Void>) () -> processImpl(contract, context, invocationType, invocationParams));
        } else {
            processImpl(contract, context, invocationType, invocationParams);
        }
    }

    private <T extends AbstractContractContext> Void processImpl(Contract contract, T context, INVOCATION_TYPE invocationType, JO invocationParams) {
        contract.injectParams(
                context.getContractRunnerConfigParams(context.getContractName()),
                contract.getContractParams(),
                invocationParams);
        switch (invocationType) {
            case TRANSACTION:
                if (!(context instanceof TransactionContext)) {
                    throw new IllegalStateException();
                }
                contract.processTransaction((TransactionContext) context);
                break;
            case BLOCK:
                if (!(context instanceof BlockContext)) {
                    throw new IllegalStateException();
                }
                contract.processBlock((BlockContext) context);
                break;
            case REQUEST:
                if (!(context instanceof RequestContext)) {
                    throw new IllegalStateException();
                }
                try {
                    contract.processRequest((RequestContext) context);
                } catch (NxtException e) {
                    throw new IllegalStateException(e);
                }
                break;
            case VOUCHER:
                if (!(context instanceof VoucherContext)) {
                    throw new IllegalStateException();
                }
                contract.processVoucher((VoucherContext) context);
                break;
        }
        return null;
    }

    private void processConfirmed(List<? extends Transaction> transactions) {
        if (isSuspendContractRunnerExecution()) {
            return;
        }
        for (Transaction transaction : transactions) {
            if (transaction.getType() == ChildBlockFxtTransactionType.INSTANCE) {
                // We process phased transactions only when released and child chain block transactions cannot trigger a contract
                continue;
            }
            if (transaction.isPhased() && !config.isValidator()) {
                // At the moment a contract cannot be triggered by phased transaction (may change in the future)
                // However, contract validatioon can be triggered by a phased transaction submitted by another contract account under its control
                continue;
            }
            try {
                JSONObject contractResponse = processTransaction(transaction, true, config.isValidator());
                if (contractResponse != null) {
                    Logger.logInfoMessage("ContractRunner response: " + contractResponse.toJSONString());
                }
            } catch (Exception e) {
                ChainTransactionId txid = new ChainTransactionId(transaction.getChain().getId(), transaction.getFullHash());
                Logger.logErrorMessage("ContractRunner error for transaction " + txid + ": " + e.getMessage(), e);
            }
        }
    }

    private void processReleasedPhased(List<? extends Transaction> transactions) {
        if (isSuspendContractRunnerExecution()) {
            return;
        }
        for (Transaction transaction : transactions) {
            try {
                JSONObject contractResponse = processTransaction(transaction, true, config.isValidator());
                if (contractResponse != null) {
                    Logger.logInfoMessage("ContractRunner response: " + contractResponse.toJSONString());
                }
            } catch (Exception e) {
                ChainTransactionId txid = new ChainTransactionId(transaction.getChain().getId(), transaction.getFullHash());
                Logger.logErrorMessage("ContractRunner error for phased transaction " + txid + ": " + e.getMessage(), e);
            }
        }
    }

    private void contractAdded(ContractReference contractReference) {
        if (contractReference.getAccountId() == config.getAccountId()) {
            addedContractReferences.put(contractReference.getContractName(), contractReference);
        }
    }

    private void contractDeleted(ContractReference contractReference) {
        if (contractReference.getAccountId() == config.getAccountId()) {
            deletedContractReferences.put(contractReference.getContractName(), contractReference);
        }
    }

    private void processBlock(Block block) {
        try {
            addedContractReferences.forEach((contractName, contractReference) ->
                    ContractLoader.loadContract(contractReference, supportedContracts, supportedContractReferences));
            deletedContractReferences.forEach((contractName, contractReference) -> {
                supportedContracts.remove(contractName);
                supportedContractReferences.remove(contractName);
            });
            addedContractReferences.clear();
            deletedContractReferences.clear();

            if (isSuspendContractRunnerExecution()) {
                return;
            }
            supportedContracts.keySet().forEach(contractName -> {
                try {
                    processBlockContract(block, contractName, true, false, null);
                } catch (Throwable t) {
                    throw new IllegalStateException("Contract " + contractName, t);
                }
            });
        } catch (Throwable t) {
            Logger.logErrorMessage("ContractRunner error: " + t.getMessage(), t);
        }
    }

    JSONObject processBlockContract(Block block, String contractName, boolean isApply, boolean isValidator, Transaction transactionToApprove) {
        Contract contract = supportedContracts.get(contractName);
        if (contract == null) {
            return generateErrorResponse(1003, "Contract is not supported %s", contractName);
        }
        BlockContext context = new BlockContext(block, config, contractName);
        process(contract, context, BLOCK);
        JO jo = context.getResponse();
        if (jo == null) {
            return null;
        }
        JSONObject contractResponse = jo.toJSONObject();
        Logger.logInfoMessage("ContractRunner processBlock at height " + block.getHeight() + " response: " + contractResponse.toJSONString());
        if (!isApply) {
            String message = "Contract simulator does not submit transactions";
            Logger.logInfoMessage(message);
            contractResponse.put("status", message);
            return contractResponse;
        }
        if (contractResponse.get("transactions") == null) {
            return null;
        }
        JSONArray transactions = ((JA) contractResponse.get("transactions")).toJSONArray();
        if (!isValidator) {
            submitContractTransactions(transactions);
        } else {
            return approveTransaction(transactionToApprove, transactions);
        }
        return null;
    }

    JSONObject processTransaction(Transaction contractOrTriggerTransaction, boolean isApply, boolean validator) {
        Logger.logInfoMessage(String.format("ContractRunner Process transaction %d:%s", contractOrTriggerTransaction.getChain().getId(), Convert.toHexString(contractOrTriggerTransaction.getFullHash())));

        // Parse the trigger message attachment for the transaction and extract the contract information
        JSONObject messageJson = parsePrunableMessage(contractOrTriggerTransaction);
        if (messageJson.get("errorDescription") != null) {
            return messageJson;
        }
        String contractName = messageJson.get("contract") != null ? (String) messageJson.get("contract") : null;
        EventSource source = messageJson.get("source") != null ? EventSource.valueOf((String) messageJson.get("source")) : EventSource.NONE;
        String seed = messageJson.get("seed") != null ? (String) messageJson.get("seed") : null;
        ChainTransactionId referencedTransactionId = null;
        if (contractOrTriggerTransaction instanceof ChildTransaction) {
            referencedTransactionId = ((ChildTransaction) contractOrTriggerTransaction).getReferencedTransactionId();
        }
        Transaction triggerTransaction;
        boolean isValidator;
        if (referencedTransactionId != null && source.isTransaction() && validator) {
            // Validate a transaction generated by a contract by loading the trigger transaction and the original contract
            // and run it again to receive the same results
            triggerTransaction = Nxt.getBlockchain().getTransaction(referencedTransactionId.getChain(), referencedTransactionId.getFullHash());
            messageJson = parsePrunableMessage(triggerTransaction); // This points to the contract which the trigger transaction has triggered
            contractName = messageJson.get("contract") != null ? (String) messageJson.get("contract") : null;
            seed = messageJson.get("seed") != null ? (String) messageJson.get("seed") : null;
            isValidator = true;
        } else if (source.isTransaction() && validator) {
            // To Validate parent chain transactions submitted by a contract we cannot use referenced transaction so we point to the trigger transaction
            // using an attached message
            ChainTransactionId triggerTransactionId = ChainTransactionId.fromStringId((String) messageJson.get("trigger"));
            if (triggerTransactionId == null) {
                return generateErrorResponse(1000, "Cannot parse trigger transaction %s", messageJson.toJSONString());
            }
            triggerTransaction = Nxt.getBlockchain().getTransaction(triggerTransactionId.getChain(), triggerTransactionId.getFullHash());
            messageJson = parsePrunableMessage(triggerTransaction);
            contractName = messageJson.get("contract") != null ? (String) messageJson.get("contract") : null;
            isValidator = true;
        } else if (source.isBlock() && validator) {
            // To validate contract transactions submitted by process block we determine the transaction execution height
            // then re-process the block at this height to validate that the contract generates the same transaction
            int height;
            if (contractOrTriggerTransaction.isPhased()) {
                // TODO: test
                PhasingPollHome.PhasingPollResult phasingPollResult = PhasingPollHome.getResult(contractOrTriggerTransaction);
                if (phasingPollResult != null) {
                    if (!phasingPollResult.isApproved()) {
                        return generateErrorResponse(1000, "Validation failed - phased transaction submitted by contract wss not approved %s", Convert.toHexString(contractOrTriggerTransaction.getFullHash()));
                    }
                    height = phasingPollResult.getHeight();
                } else {
                    return generateErrorResponse(1000, "Validation failed - phased transaction submitted by contract not executed yet %s", Convert.toHexString(contractOrTriggerTransaction.getFullHash()));
                }
            } else {
                height = contractOrTriggerTransaction.getECBlockHeight();
            }

            Block block = Nxt.getBlockchain().getBlockAtHeight(height);
            contractName = messageJson.get("submittedBy") != null ? (String) messageJson.get("submittedBy") : null;
            if (contractName == null) {
                return generateErrorResponse(1000, "Cannot trigger contract, contract id not specified %s", messageJson.toJSONString());
            }
            return processBlockContract(block, contractName, true, true, contractOrTriggerTransaction);
        } else {
            // This is main use case when we receive a new trigger transaction
            triggerTransaction = contractOrTriggerTransaction;
            isValidator = false;
        }
        if (contractName == null) {
            return generateErrorResponse(1000, "ContractRunner trigger %s did not specify contract name", Convert.toHexString(triggerTransaction.getFullHash()));
        }
        Contract contract = supportedContracts.get(contractName);
        if (contract == null) {
            return generateErrorResponse(1000, "Contract %s is not supported", contractName);
        }
        JO contractInvocationParamsJson;
        if (messageJson.get("params") != null) {
            contractInvocationParamsJson = new JO(messageJson.get("params"));
        } else {
            contractInvocationParamsJson = new JO();
        }
        Logger.logInfoMessage(String.format("Executing contract name %s class %s", contractName, contract.getClass().getCanonicalName()));
        TransactionContext context = new TransactionContext(triggerTransaction, config, contractInvocationParamsJson, contractName, seed);
        process(contract, context, TRANSACTION);
        JO jo = context.getResponse();
        JSONObject contractResponse = jo.toJSONObject();
        if (!isApply) {
            String message = "Contract simulator does not submit transactions";
            Logger.logInfoMessage(message);
            contractResponse.put("status", message);
            return contractResponse;
        }
        JSONArray transactions;
        if (contractResponse.get("transactions") != null) {
            // Transactions generated by the contract but not submitted yet
            transactions = ((JA) contractResponse.get("transactions")).toJSONArray();
        } else {
            transactions = new JSONArray();
        }
        if (!isValidator) {
            submitContractTransactions(transactions);
            contractResponse.put("status", "Transaction submitted");
            return contractResponse;
        } else {
            contractResponse = approveTransaction(contractOrTriggerTransaction, transactions);
            contractResponse.put("status", "Approval transaction submitted");
            return contractResponse;
        }
    }

    /**
     * Let a runner validate a transaction submitted by another contract runner instance.
     * Given a transaction generated by a contract previously and the transactions generated by the current invocation
     * If the transaction to approve is equal to one of the generated transactions, submit an approval transaction
     *
     * @param transactionToApprove the transaction we are trying to approve
     * @param contractTransactions transactions generated by the current contract invocation
     * @return the approve transaction JSON or an error message if not approved
     */
    private JSONObject approveTransaction(Transaction transactionToApprove, JSONArray contractTransactions) {
        // First we normalize the data of the transaction objects so that we can compare them
        JO expectedTransactionJSON = new JO(JSONData.unconfirmedTransaction(transactionToApprove));
        TransactionResponse expectedTransaction = TransactionResponse.create(expectedTransactionJSON);
        for (Object transactionObject : contractTransactions) {
            JO actualTransactionJSON = new JO(new JO(transactionObject).getJo("transactionJSON"));
            TransactionResponse actualTransaction = TransactionResponse.create(actualTransactionJSON);
            if (actualTransaction.equals(expectedTransaction)) {
                if (transactionToApprove.getChain() == FxtChain.FXT) {
                    return generateErrorResponse(1000, "Found a match but cannot submit approval to a parent chain transaction");
                }
                // If we found a match, we submit approval transaction to the original transaction
                String validatorSecretPhrase = config.getValidatorSecretPhrase();
                if (validatorSecretPhrase == null) {
                    return generateErrorResponse(1000, "Cannot approve transaction, validatorSecretPhrase not specified");
                }
                expectedTransactionJSON = new JO(JSONData.unconfirmedTransaction(transactionToApprove));
                APICall.Builder builder = new APICall.Builder("approveTransaction").
                        secretPhrase(config.getSecretPhrase()).
                        param("phasedTransaction", expectedTransactionJSON.getInt("chain") + ":" + expectedTransactionJSON.getString("fullHash")).
                        secretPhrase(validatorSecretPhrase);
                int chainId = (int) expectedTransactionJSON.get("chain");
                if (Chain.getChain(chainId) instanceof ChildChain) {
                    builder.param("feeRateNQTPerFXT", config.getFeeRateNQTPerFXT(chainId));
                }
                return builder.build().invoke();
            } else {
                Logger.logInfoMessage("Transactions differ");
                Logger.logInfoMessage("Expected Transaction " + expectedTransactionJSON.toJSONString());
                Logger.logInfoMessage("Actual   Transaction " + actualTransactionJSON.toJSONString());
            }
        }
        return generateErrorResponse(1000, "Cannot approve contract transaction %s chain %s", expectedTransactionJSON.get("fullHash"), expectedTransactionJSON.get("chain"));
    }

    void submitContractTransactions(JSONArray transactions) {
        for (Object transactionObject : transactions) {
            JO transaction = new JO(transactionObject);
            if (!transaction.isExist("transactionJSON")) {
                Logger.logErrorMessage(String.format("Error %s %s in transaction submitted by contract",
                        transaction.get("errorCode"), transaction.get("errorDescription")));
                continue;
            }
            JO transactionJSON = transaction.getJo("transactionJSON");
            APICall.Builder builder;
            APICall apiCall;
            if (!transactionJSON.isExist("signature")) {
                builder = new APICall.Builder("signTransaction").
                        secretPhrase(config.getSecretPhrase()).
                        param("unsignedTransactionJSON", transactionJSON.toJSONString()).
                        param("validate", "true");
                apiCall = builder.build();
                JO signTransactionResponse = new JO(apiCall.invoke());
                if (signTransactionResponse.isExist("errorCode")) {
                    Logger.logErrorMessage(String.format("Error signing transaction %s chain %d message %s",
                            transactionJSON.getString("fullHash"), transactionJSON.getLong("chain"), signTransactionResponse.getString("errorDescription")));
                    continue;
                }
                transactionJSON = new JO(signTransactionResponse.get("transactionJSON"));
            }
            builder = new APICall.Builder("broadcastTransaction").
                    param("transactionJSON", transactionJSON.toJSONString());
            apiCall = builder.build();
            JO broadcastTransactionResponse = new JO(apiCall.invoke());
            if (broadcastTransactionResponse.get("errorCode") != null) {
                Logger.logErrorMessage(String.format("Error broadcasting transaction %s chain %d message %s",
                        transactionJSON.getString("fullHash"), transactionJSON.getLong("chain"), broadcastTransactionResponse.getString("errorDescription")));
            }
        }
    }

    private boolean isSuspendContractRunnerExecution() {
        return Nxt.getBlockchain().getLastBlockTimestamp() < Nxt.getEpochTime() - config.getCatchUpInterval();
    }

    JSONObject generateErrorResponse(int code, String message, Object... params) {
        message = String.format(message, params);
        Logger.logInfoMessage(message);
        JSONObject response = new JSONObject();
        response.put("errorCode", code);
        response.put("errorDescription", message);
        return response;
    }

    /**
     * Load JSON data from transaction attached prunable message, plain or encrypted
     *
     * @param transaction the transaction
     * @return the attached message in JSON format
     */
    private JSONObject parsePrunableMessage(Transaction transaction) {
        PrunablePlainMessageAppendix plainAppendix = (PrunablePlainMessageAppendix) transaction.getAppendages().stream().filter(a -> a instanceof PrunablePlainMessageAppendix).findFirst().orElse(null);
        PrunableEncryptedMessageAppendix encryptedAppendix = (PrunableEncryptedMessageAppendix) transaction.getAppendages().stream().filter(a -> a instanceof PrunableEncryptedMessageAppendix).findFirst().orElse(null);
        String messageText;
        if (plainAppendix != null) {
            messageText = Convert.toString(plainAppendix.getMessage(), plainAppendix.isText());
        } else if (encryptedAppendix != null) {
            if (transaction.getRecipientId() != config.getAccountId()) {
                return generateErrorResponse(1000, "Cannot decrypt attached message, contract account %s is not the recipient %s of attached message", config.getAccountRs(), Convert.rsAccount(transaction.getRecipientId()));
            } else if (config.isValidator()) {
                // TODO if this becomes an important limitation perhaps we can rely on a shared key in this case
                return generateErrorResponse(1000, "Cannot decrypt attached message, validator cannot decrypt encrypted trigger message");
            }
            messageText = Convert.toString(Account.decryptFrom(transaction.getSenderPublicKey(), encryptedAppendix.getEncryptedData(), config.getSecretPhrase(), encryptedAppendix.isCompressed()), encryptedAppendix.isText());
        } else {
            return generateErrorResponse(1000, "Transaction %s of chain %s does not trigger a contract", Convert.toHexString(transaction.getFullHash()), transaction.getChain());
        }
        try {
            Object messageObject = JSONValue.parseWithException(messageText);
            if (!(messageObject instanceof JSONObject)) {
                return generateErrorResponse(1000, "trigger message does not represent a JSON object");
            }
            return (JSONObject) messageObject;
        } catch (Exception e) {
            return generateErrorResponse(1000, "cannot parse attached message, probably not a trigger transaction " + e);
        }
    }

    @Override
    public void shutdown() {
        Logger.logInfoMessage("ContractRunner shutdown");
    }

    public void reset() {
        supportedContracts.clear();
        supportedContractReferences.clear();
    }

    @Override
    public Map<String, APIServlet.APIRequestHandler> getAPIRequests() {
        return apiRequests;
    }

    @Override
    public Contract getContract(String name) {
        return supportedContracts.get(name);
    }

    ContractRunnerConfig getConfig() {
        return config;
    }

    Map<String, Contract> getSupportedContracts() {
        return supportedContracts;
    }

    Map<String, ContractReference> getSupportedContractReferences() {
        return supportedContractReferences;
    }

}
